
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "libmikmod-3.1.12/include/mikmod_build.h"

static FILE *FileHandle;

static MODULE *pMOD;

extern u32 ModSoundDriver_SamplesPerFlame;
u32 ModSoundDriver_SamplesPerFlame;
extern u32 *ModSoundDriver_pSamples;
u32 *ModSoundDriver_pSamples;

static const float TimeSecDiv=5;

// ------------------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("libmikmod, version 3.1.12 Copyright (C) 1989 by Jef Poskanzer.\n");
  printf("LibSndMOD by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  u32 SampleRate=44100;
  
  FileHandle=fopen(pFilename,"r");
  if(FileHandle==NULL){
    snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
    return(false);
  }
  
  ModSoundDriver_SamplesPerFlame=512;
  ModSoundDriver_pSamples=NULL;
  
  MikMod_RegisterAllDrivers();
  MikMod_RegisterAllLoaders();
  
  /* initialize the library */
  md_mode|=DMODE_16BITS|DMODE_STEREO|DMODE_HQMIXER|DMODE_SURROUND|DMODE_INTERP|DMODE_SOFT_MUSIC;
  md_device=1;
  if(MikMod_Init(NULL)){
    snprintf(pState->ErrorMessage,256,"Could not initialize sound, reason: %s", MikMod_strerror(MikMod_errno));
    return(false);
  }
  
  /* load module */
  pMOD = Player_Load((char*)pFilename,64,0);
  if(!pMOD){
    snprintf(pState->ErrorMessage,256,"Could not load module, reason: %s", MikMod_strerror(MikMod_errno));
    MikMod_Exit();
    return(false);
  }
  
  Player_Start(pMOD);
  
  pState->SampleRate=SampleRate;
  pState->SamplesPerFlame=ModSoundDriver_SamplesPerFlame;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=pMOD->numpos*TimeSecDiv;
  
  {
    const char *pTitle=pMOD->songname;
    if(pTitle!=NULL) pState->pTitleW=Unicode_AllocateCopyFromUTF8(pTitle);
  }
  
  LibSnd_DebugOut("MOD decoder initialized.\n");
  
  return(true);
}

static bool isComment(void)
{
  if(pMOD->comment==NULL) return(false);
  
  char *modtype=pMOD->modtype;
  
  if(strncmp(modtype,"ImpulseTracker",strlen("ImpulseTracker"))==0) return(true);
  if(strncmp(modtype,"Compressed ImpulseTracker",strlen("Compressed ImpulseTracker"))==0) return(true);
  if(strncmp(modtype,"MTM",strlen("MTM"))==0) return(true);
  
  return(false);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  {
    const char *pmodtype=pMOD->modtype;
    if(pmodtype==NULL) pmodtype="";
    printf("%s inst: %d, smpl: %d.\n",pmodtype,pMOD->numins,pMOD->numsmp);
  }
  
  {
    const char *ptitle=pMOD->songname;
    if(ptitle==NULL) ptitle="no title";
    printf("Title: %s\n",ptitle);
  }
  
  if(isComment()==true) printf("%s\n",pMOD->comment);
}

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  Player_SetPosition(sec/TimeSecDiv);
  pState->CurrentSec=sec;
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(Player_Active()==false) pState->isEnd=true;
  
  if(pState->isEnd==true) return(0);
  
  ModSoundDriver_pSamples=pBufLR;
  MikMod_Update();
  ModSoundDriver_pSamples=NULL;
  
  pState->CurrentSec=pMOD->sngpos*TimeSecDiv;
  
  return(ModSoundDriver_SamplesPerFlame);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pMOD!=NULL){
    Player_Stop();
    Player_Free(pMOD); pMOD=NULL;
  }
  
  MikMod_Exit();
  
  if(FileHandle!=NULL){
    fclose(FileHandle); FileHandle=NULL;
  }
}

TLibSnd_Interface* LibSndMOD_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

