
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pspuser.h>

#include "euc2unicode.h"

#include "LibSnd_Const_Global.h"
#include "LibSnd_Const_Internal.h"

static TLibSndConst_State LibSndInt_State;

// ------------------------------------------------------------------------------------

#include "ttalib-hwplayer-1.2/ttalib.h"

static FILE *FileHandle;

#define MAX_BSIZE (MAX_BPS>>3)
static tta_info info; // currently playing file info

// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------

static void LibSndInt_ShowLicense(void)
{
  printf("LibSndTTA by Moonlight.\n");
  printf("\n");
}

static bool LibSndInt_Open(const char *pFilename,const u32 TrackNum)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  FileHandle=fopen(pFilename,"r");
  if(FileHandle==NULL){
    snprintf(pState->ErrorMessage,256,"File not found. [%s]",pFilename);
    return(false);
  }
  
  if(open_tta_file(FileHandle,&info,0)<0){
    snprintf(pState->ErrorMessage,256,"TTA Decoder Error - %s",get_error_str(info.STATE));
    close_tta_file(&info);
    return(false);
  }
  
  if(player_init(&info)<0){
    snprintf(pState->ErrorMessage,256,"TTA Decoder Error - %s",get_error_str(info.STATE));
    close_tta_file(&info);
    return(false);
  }
  
  if((info.BSIZE!=1)&&(info.BSIZE!=2)){
    snprintf(pState->ErrorMessage,256,"Not support %dbits format.",info.BSIZE*8);
    close_tta_file(&info);
    return(false);
  }
  
  pState->SampleRate=info.SAMPLERATE;
  pState->SamplesPerFlame=PCM_BUFFER_LENGTH;
  pState->isEnd=false;
  pState->CurrentSec=0;
  pState->TotalSec=info.LENGTH;
  
  {
    char msg[1024];
    {
      char *pmsg=msg;
      pmsg[0]=0;
      if(info.ID3.title!=NULL) pmsg+=snprintf(pmsg,256,"Title:%s",info.ID3.title);
      if(info.ID3.artist!=NULL) pmsg+=snprintf(pmsg,256,"/ Artist:%s",info.ID3.artist);
      if(info.ID3.album!=NULL) pmsg+=snprintf(pmsg,256,"/ Album:%s",info.ID3.album);
      if(info.ID3.track!=NULL) pmsg+=snprintf(pmsg,256,"/ Track:%s",info.ID3.track);
      if(info.ID3.year!=NULL) pmsg+=snprintf(pmsg,256,"/ Year:%s",info.ID3.year);
      if(info.ID3.genre!=NULL) pmsg+=snprintf(pmsg,256,"/ Genre:%s",info.ID3.genre);
      if(info.ID3.comment!=NULL) pmsg+=snprintf(pmsg,256,"/ Comment:%s",info.ID3.comment);
    }
    pState->pTitleW=EUC2Unicode_Convert(msg);
  }
  
  LibSnd_DebugOut("TTA decoder initialized.\n");
  
  return(true);
}

static void LibSndInt_ShowStateEx(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  // Show file info
  printf("File size: [%d]\n",info.FILESIZE);
  printf("Number of channels: [%d]\n",(int)info.NCH);
  printf("Bits per sample: [%d]\n",(int)info.BPS);
  printf("Audio format: [%d]\n",(int)info.FORMAT);
  printf("Samplerate: [%d]\n",info.SAMPLERATE);
  printf("Sample/bytes: [%d]\n",info.BSIZE);
  printf("Playback time: [%d:%02d]\n",info.LENGTH/60,info.LENGTH%60);
  printf("Average bitrate: [%d Kbps]\n",info.BITRATE);
  printf("Compression ratio: [%.2f]\n",info.COMPRESS);
  
  printf("\nTitle: [%s]\n",info.ID3.title);
  printf("Artist: [%s]\n",info.ID3.artist);
  printf("Album: [%s]\n",info.ID3.album);
  printf("Track: [%s]\n",info.ID3.track);
  printf("Year: [%s]\n",info.ID3.year);
  printf("Genre: [%s]\n",info.ID3.genre);
  printf("Comment: [%s]\n\n",info.ID3.comment);
}

static void LibSndInt_Seek(float sec)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(set_position(sec*1000/SEEK_STEP)!=0){
    printf("TTA seek error: %s\n", get_error_str(info.STATE));
    return;
  }
  pState->CurrentSec=sec;
}

static u32 LibSndInt_Update(u32 *pBufLR)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  if(pState->isEnd==true) return(0);
  
  u8 sample_buffer[PCM_BUFFER_LENGTH * MAX_BSIZE * MAX_NCH];
  
  u32 SamplesCount=get_samples(sample_buffer);
  if(SamplesCount<0){
    snprintf(pState->ErrorMessage,256,"TTA Decoder Error - %s\n",get_error_str(info.STATE));
    return(0);
  }
  
  if(SamplesCount==0){
    pState->isEnd=true;
    return(0);
  }
  
  u32 nch=info.NCH;
  
  switch(info.BSIZE){
    case 1: { // bps 8
      s8 *psrc=(s8*)sample_buffer;
      for(u32 idx=0;idx<SamplesCount;idx++){
        s32 sl,sr;
        sl=psrc[0]<<8;
        if(nch==1){
          sr=sl;
          }else{
          sr=psrc[1]<<8;
        }
        *pBufLR++=(sl&0xffff)|(sr<<16);
        psrc+=nch;
      }
    } break;
    case 2: { // bps 16
      s16 *psrc=(s16*)sample_buffer;
      for(u32 idx=0;idx<SamplesCount;idx++){
        s32 sl,sr;
        sl=psrc[0];
        if(nch==1){
          sr=sl;
          }else{
          sr=psrc[1];
        }
        *pBufLR++=(sl&0xffff)|(sr<<16);
        psrc+=nch;
      }
    } break;
  }
  
  pState->CurrentSec+=(float)SamplesCount/pState->SampleRate;
  
  return(SamplesCount);
}

static void LibSndInt_Close(void)
{
  TLibSndConst_State *pState=&LibSndInt_State;
  
  player_stop ();
  close_tta_file (&info);
}

TLibSnd_Interface* LibSndTTA_GetInterface(void)
{
  static TLibSnd_Interface res={
    LibSndInt_ShowLicense,
    LibSndInt_Open,
    LibSndInt_ShowStateEx,
    LibSndInt_Seek,
    LibSndInt_Update,
    LibSndInt_Close,
    &LibSndInt_State,
  };
  return(&res);
}

