
static const u32 Page_BGColor32=0xff554433;

static const u32 BlkXSizeFull=512/2,BlkYSizeFull=512/2;
//   fit  1.0
// 1 37ms 30ms
// 2 15ms 31ms
// 4 17ms 32ms
// 8 24ms 25ms

typedef struct {
  u32 BlkXSize,BlkYSize;
  u32 **ppBlkBufs;
} TPage_MipMap;
static const u32 MipMapCount=3;

typedef struct {
  CFont *pCFont;
  u32 Width,Height;
  u32 BlkXCnt,BlkYCnt;
  TPage_MipMap MipMaps[MipMapCount];
  float ViewLeft,ViewTop;
  float ViewRatio;
} TPage;

static TPage Page;

static const u32 msgbufsize=128;
static char msgbuf[msgbufsize];

static const float MaxRatio=7;

static void Page_Load_Error(const char *pmsg)
{
  TPage *ppg=&Page;
  
  ppg->pCFont->DrawTextA_White(pGUViewBuf,ScreenLineSize,8,ScreenHeight-48,pmsg);
  SystemHalt();
}

static void Page_SetRatio_FitToScreen(void)
{
  TPage *ppg=&Page;
  
  float rx=(float)ScreenWidth/ppg->Width,ry=(float)ScreenHeight/ppg->Height;
  if(rx<ry){
    ppg->ViewRatio=rx;
    }else{
    ppg->ViewRatio=ry;
  }
  
  if(MaxRatio<ppg->ViewRatio) ppg->ViewRatio=MaxRatio;
}

static void Page_MoveToInsideScreen(void)
{
  TPage *ppg=&Page;
  
  float scrw=ScreenWidth/ppg->ViewRatio;
  float scrh=ScreenHeight/ppg->ViewRatio;
  if((ppg->Width-scrw)<ppg->ViewLeft) ppg->ViewLeft=ppg->Width-scrw;
  if((ppg->Height-scrh)<ppg->ViewTop) ppg->ViewTop=ppg->Height-scrh;
  if(ppg->ViewLeft<0) ppg->ViewLeft=0;
  if(ppg->ViewTop<0) ppg->ViewTop=0;
}

static void Page_ChangeViewRatio(float r)
{
  TPage *ppg=&Page;
  
  float cx=ppg->ViewLeft+(ScreenWidth/2/ppg->ViewRatio);
  float cy=ppg->ViewTop+(ScreenHeight/2/ppg->ViewRatio);
  
  if(r<0.01) r=0.01;
  if(MaxRatio<r) r=MaxRatio;
  ppg->ViewRatio=r;
  
  ppg->ViewLeft=cx-(ScreenWidth/2/ppg->ViewRatio);
  ppg->ViewTop=cy-(ScreenHeight/2/ppg->ViewRatio);
}

static u32 PrgBarTitlePosY;

static void Page_Load_ShowPrgBar(const char *pTitle)
{
  TPage *ppg=&Page;
  
  {
    if(PrgBarTitlePosY==0) PrgBarTitlePosY=(ScreenHeight/2)+32;
    const u32 TextHeight=ppg->pCFont->GetTextHeight();
    const u32 x=8,y=PrgBarTitlePosY;
    PrgBarTitlePosY+=TextHeight+8;
    ppg->pCFont->DrawTextA_White(pGUViewBuf,ScreenLineSize,x,y,pTitle);
  }
  
  const u32 PrgTop=8+ppg->pCFont->GetTextHeight()*7;
  
  u32 *pbuf=pGUViewBuf;
  pbuf+=PrgTop*ScreenLineSize;
  
  {
    for(u32 x=8;x<ScreenWidth-8;x++){
      pbuf[x]=0x00ffffff;
    }
    pbuf+=ScreenLineSize;
  }
  
  {
    for(u32 y=0;y<32;y++){
      pbuf[7]=0x00ffffff;
      for(u32 x=8;x<ScreenWidth-8;x++){
        pbuf[x]=0x00400000;
      }
      pbuf[ScreenWidth-8]=0x00ffffff;
      pbuf+=ScreenLineSize;
    }
  }
  
  {
    for(u32 x=8;x<ScreenWidth-8;x++){
      pbuf[x]=0x00ffffff;
    }
    pbuf+=ScreenLineSize;
  }
}

static void Page_Load_RefreshPrgBar(u32 pos,u32 max)
{
  TPage *ppg=&Page;
  
  const u32 PrgTop=8+ppg->pCFont->GetTextHeight()*7;
  
  u32 *pbuf=pGUViewBuf;
  pbuf+=PrgTop*ScreenLineSize;
  
  pbuf+=ScreenLineSize;
  
  const u32 barw=pos*(ScreenWidth-8-8)/max;
  for(u32 y=0;y<32;y++){
    for(u32 x=0;x<barw;x++){
      pbuf[8+x]=0x00ff4040;
    }
    pbuf+=ScreenLineSize;
  }
}

static void Page_Load(const char *pfn)
{
  TPage *ppg=&Page;
  
  ppg->pCFont=pCFont16;
  
  if(LibImg_Start(pfn)==false){
    snprintf(msgbuf,msgbufsize,"LibImg open error.");
    Page_Load_Error(msgbuf);
  }
  
  const TLibImgConst_State *pState=LibImg_GetState();
  
  {
    snprintf(msgbuf,msgbufsize,"Image info: %dx%dpixels.",pState->Width,pState->Height);
    ppg->pCFont->DrawTextA_White(pGUViewBuf,ScreenLineSize,8,8+ppg->pCFont->GetTextHeight()*3,msgbuf);
  }
  
  ppg->Width=pState->Width;
  ppg->Height=pState->Height;
  
  ppg->BlkXCnt=(ppg->Width+(BlkXSizeFull-1))/BlkXSizeFull;
  ppg->BlkYCnt=(ppg->Height+(BlkYSizeFull-1))/BlkYSizeFull;
  printf("Blocks count: %dx%d.\n",ppg->BlkXCnt,ppg->BlkYCnt);
  for(u32 midx=0;midx<MipMapCount;midx++){
    TPage_MipMap *pmm=&ppg->MipMaps[midx];
    
    {
      u32 BlkXSize=BlkXSizeFull;
      u32 BlkYSize=BlkYSizeFull;
      for(u32 idx=0;idx<midx;idx++){
        BlkXSize=BlkXSize/2;
        BlkYSize=BlkYSize/2;
      }
      printf("MipMapSize:%d,%d,%d.\n",midx,BlkXSize,BlkYSize);
      pmm->BlkXSize=BlkXSize;
      pmm->BlkYSize=BlkYSize;
    }
    
    pmm->ppBlkBufs=(u32**)malloc(ppg->BlkXCnt*ppg->BlkYCnt*sizeof(u32*));
    if(pmm->ppBlkBufs==NULL){
      snprintf(msgbuf,msgbufsize,"%s:%d Memory overflow.",__FILE__,__LINE__);
      Page_Load_Error(msgbuf);
    }
    
    for(u32 by=0;by<ppg->BlkYCnt;by++){
      for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
        u32 BlkIdx=(by*ppg->BlkXCnt)+bx;
        pmm->ppBlkBufs[BlkIdx]=(u32*)malloc(pmm->BlkXSize*pmm->BlkYSize*sizeof(u32));
        if(pmm->ppBlkBufs[BlkIdx]==NULL){
          snprintf(msgbuf,msgbufsize,"%s:%d Memory overflow.",__FILE__,__LINE__);
          Page_Load_Error(msgbuf);
        }
        u32 *pbuf=pmm->ppBlkBufs[BlkIdx];
        for(u32 y=0;y<pmm->BlkXSize*pmm->BlkYSize;y++){
          *pbuf++=Page_BGColor32;
        }
      }
    }
  }
  
  PrgBarTitlePosY=0;
  
  {
    Page_Load_ShowPrgBar("Decoding image stream...");
    
    const u32 LineBufSize=ppg->MipMaps[0].BlkXSize*ppg->BlkXCnt;
    u32 *pLineBuf=(u32*)malloc(LineBufSize*sizeof(u32));
    if(pLineBuf==NULL){
      snprintf(msgbuf,msgbufsize,"%s:%d Memory overflow.",__FILE__,__LINE__);
      Page_Load_Error(msgbuf);
    }
    
    for(u32 y=0;y<ppg->Height;y++){
      for(u32 idx=0;idx<LineBufSize;idx++){
        pLineBuf[idx]=Page_BGColor32;
      }
      if((y&7)==0) Page_Load_RefreshPrgBar(y,ppg->Height);
      if(LibImg_Decode_RGBA8888(pLineBuf,0)==false) printf("Decode error. (%d)\n",y);
      TPage_MipMap *pmm=&ppg->MipMaps[0];
      u32 BlkLeftIndex=(y/pmm->BlkYSize)*ppg->BlkXCnt;
      for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
        u32 *psrc=&pLineBuf[bx*pmm->BlkXSize];
        u32 *pdst=pmm->ppBlkBufs[BlkLeftIndex+bx];
        pdst+=(y%pmm->BlkYSize)*pmm->BlkXSize;
        for(u32 x=0;x<pmm->BlkXSize;x++){
          *pdst++=*psrc++;
        }
      }
    }
    
    if(pLineBuf!=NULL){
      free(pLineBuf); pLineBuf=NULL;
    }
  }
  
  printf("Decoded.\n");
  
  Page_Load_ShowPrgBar("MipMaps and Swizzle filter...");
  
  for(u32 midx=0;midx<MipMapCount-1;midx++){
    Page_Load_RefreshPrgBar(midx,(MipMapCount-1)*2);
    TPage_MipMap *pmmsrc=&ppg->MipMaps[midx+0];
    TPage_MipMap *pmmdst=&ppg->MipMaps[midx+1];
    
    for(u32 by=0;by<ppg->BlkYCnt;by++){
      for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
        u32 BlkIdx=(by*ppg->BlkXCnt)+bx;
        u32 *psrcbuf=pmmsrc->ppBlkBufs[BlkIdx];
        u32 *pdstbuf=pmmdst->ppBlkBufs[BlkIdx];
        const u32 sw=pmmsrc->BlkXSize,sh=pmmsrc->BlkYSize;
        const u32 dw=pmmdst->BlkXSize,dh=pmmdst->BlkYSize;
        if(((sw/2)!=dw)||((sh/2)!=dh)){
          snprintf(msgbuf,msgbufsize,"Illigal MipMap reduce ratio. (%d->%d),(%d->%d)",sw,dw,sh,dh);
          Page_Load_Error(msgbuf);
        }
        for(u32 dy=0;dy<dh;dy++){
          const u32 *psrc0=&psrcbuf[(dy*2+0)*sw];
          const u32 *psrc1=&psrcbuf[(dy*2+1)*sw];
          u32 *pdst=&pdstbuf[dy*dw];
          for(u32 dx=0;dx<dw;dx++){
            u32 r=0,g=0,b=0;
            for(u32 idx=0;idx<2;idx++){
              u32 c=*psrc0++;
              r+=(c>>0)&0xff;
              g+=(c>>8)&0xff;
              b+=(c>>16)&0xff;
            }
            for(u32 idx=0;idx<2;idx++){
              u32 c=*psrc1++;
              r+=(c>>0)&0xff;
              g+=(c>>8)&0xff;
              b+=(c>>16)&0xff;
            }
            r/=4; g/=4; b/=4;
            *pdst++=(0xff<<24)|(b<<16)|(g<<8)|(r<<0);
          }
        }
      }
    }
  }
  
  printf("MipMap created.\n");
  
  for(u32 midx=0;midx<MipMapCount;midx++){
    Page_Load_RefreshPrgBar(MipMapCount+midx,MipMapCount*2);
    TPage_MipMap *pmm=&ppg->MipMaps[midx];
    
    u32 *ptmp=(u32*)malloc(pmm->BlkXSize*pmm->BlkYSize*sizeof(u32));
    if(ptmp==NULL){
      snprintf(msgbuf,msgbufsize,"%s:%d Memory overflow.",__FILE__,__LINE__);
      Page_Load_Error(msgbuf);
    }
    
    for(u32 by=0;by<ppg->BlkYCnt;by++){
      for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
        u32 BlkIdx=(by*ppg->BlkXCnt)+bx;
        TextureHelper_swizzle_fast((u8*)ptmp, (u8*)pmm->ppBlkBufs[BlkIdx], pmm->BlkXSize*4,pmm->BlkYSize);
        MemCopy32CPU(ptmp,pmm->ppBlkBufs[BlkIdx],pmm->BlkXSize*pmm->BlkYSize*sizeof(u32));
      }
    }
    
    if(ptmp!=NULL){
      free(ptmp); ptmp=NULL;
    }
  }
  
  printf("Swizzled.\n");
  
  LibImg_Close();
  
  ppg->ViewLeft=0;
  ppg->ViewTop=0;
  ppg->ViewRatio=1;
  
  Page_SetRatio_FitToScreen();
}

static void Page_Free(void)
{
  TPage *ppg=&Page;
  
  ppg->pCFont=NULL;
  
  for(u32 midx=0;midx<MipMapCount;midx++){
    TPage_MipMap *pmm=&ppg->MipMaps[midx];
    
    if(pmm->ppBlkBufs!=NULL){
      for(u32 by=0;by<ppg->BlkYCnt;by++){
        for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
          u32 BlkIdx=(by*ppg->BlkXCnt)+bx;
          if(pmm->ppBlkBufs[BlkIdx]!=NULL){
            free(pmm->ppBlkBufs[BlkIdx]); pmm->ppBlkBufs[BlkIdx]=NULL;
          }
        }
      }
      free(pmm->ppBlkBufs); pmm->ppBlkBufs=NULL;
    }
  }
  
  ppg->Width=0;
  ppg->Height=0;
  
  ppg->ViewLeft=0;
  ppg->ViewTop=0;
  ppg->ViewRatio=1;
}

static void Page_Draw(void)
{
  TPage *ppg=&Page;
  
  sceGuEnable(GU_BLEND);
  sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
  
  sceGuEnable(GU_TEXTURE_2D);
  sceGuTexMode(GU_PSM_8888, MipMapCount-1, 0, GU_TRUE);
  sceGuTexFunc(GU_TFX_MODULATE, GU_TCC_RGBA);
  sceGuTexEnvColor(0x0);
  sceGuTexOffset(0.0f, 0.0f);
  sceGuTexScale(BlkXSizeFull,BlkYSizeFull);
  sceGuTexWrap(GU_CLAMP,GU_CLAMP);
  sceGuTexLevelMode(GU_TEXTURE_AUTO,0);
  sceGuTexFilter(GU_LINEAR,GU_LINEAR);
  sceGuTexFilter(GU_LINEAR_MIPMAP_LINEAR,GU_LINEAR_MIPMAP_LINEAR);
  sceGuTexFilter(GU_LINEAR_MIPMAP_NEAREST,GU_LINEAR_MIPMAP_NEAREST);
  
  typedef struct {
    float u,v;
    u32 c;
    float x,y,z;
  } TV;
  
  float r=ppg->ViewRatio;
  
  float ofsx=0,ofsy=0;
  
  {
    float scrw=ppg->Width*ppg->ViewRatio;
    float scrh=ppg->Height*ppg->ViewRatio;
    if(scrw<ScreenWidth) ofsx=(ScreenWidth-scrw)/2;
    if(scrh<ScreenHeight) ofsy=(ScreenHeight-scrh)/2;
  }
  
  if((ofsx!=0)||(ofsy!=0)){
    sceGuClearColor(Page_BGColor32);
    sceGuClear(GU_COLOR_BUFFER_BIT);
  }
  
  for(u32 by=0;by<ppg->BlkYCnt;by++){
    for(u32 bx=0;bx<ppg->BlkXCnt;bx++){
      
      for(u32 midx=0;midx<MipMapCount;midx++){
        TPage_MipMap *pmm=&ppg->MipMaps[midx];
        sceGuTexImage(midx, pmm->BlkXSize,pmm->BlkYSize,pmm->BlkXSize,pmm->ppBlkBufs[(by*ppg->BlkXCnt)+bx]);
      }
      
      TV *pV = (TV*)sceGuGetMemory(sizeof(TV) * 2);
      
      const float x=(bx*BlkXSizeFull)-ppg->ViewLeft;
      const float y=(by*BlkYSizeFull)-ppg->ViewTop;
      {
        TV *v0 = &pV[0];
        TV *v1 = &pV[1];
        
        v0->u = 0.0f;
        v0->v = 0.0f;
        v0->c = 0x00ffff00;
        v0->x = ofsx+(x*r);
        v0->y = ofsy+(y*r);
        v0->z = 0.0f;
        
        v1->u = BlkXSizeFull;
        v1->v = BlkYSizeFull;
        v1->c = 0xffffffff;
        v1->x = ofsx+((x+BlkYSizeFull)*r);
        v1->y = ofsy+((y+BlkYSizeFull)*r);
        v1->z = 0.0f;
      }
      
      sceGumDrawArray(GU_SPRITES, GU_TEXTURE_32BITF | GU_COLOR_8888 | GU_VERTEX_32BITF | GU_TRANSFORM_2D, 2, 0, pV);
    }
  }
  
  sceGuDisable(GU_TEXTURE_2D);
  
  sceKernelDcacheWritebackAll();
}


